#include "elemlist.hh"

template<typename collitemtype, class Op>
class Group3test
{
    collitemtype base;
    const Lista<Op> &list;
public:
    Group3test(collitemtype b, const Lista<Op> &l) : base(b), list(l) { }
    static unsigned groupcount() { return Op::groupcount(); }
    bool needsparm(collitemtype c) const { return list[c-base].needsparm(); }
    bool compatible(collitemtype a, collitemtype b) const { return list[a-base].compatible(list[b-base]); }
};

#include <list>
#include <vector>
#include <utility>
#include <cstdio>
#include <map>
using std::list;
using std::pair;
using std::vector;
using std::printf;
using std::make_pair;

template<typename collitemtype, class Op>
class JobList
{
    typedef typename std::set<collitemtype>::const_iterator it;
    
    collitemtype agbase, chbase, mdbase;
    const Lista<Agent> &aglist;
    const Lista<Channel> &chlist;
    const Lista<Op> &mdlist;
public:
    typedef list<pair<string, vector<collitemtype> > > joblisttype;
    joblisttype joblist;

    JobList(collitemtype a, collitemtype c, collitemtype m,
              const Lista<Agent> &ag, const Lista<Channel> &ch,
              const Lista<Op> &md) : agbase(a), chbase(c), mdbase(m),
                aglist(ag), chlist(ch), mdlist(md) { }
    void operator()
    (
      const it ag1, const it ag2,
      const it ch1, const it ch2,
      const it md1, const it md2
    )
    {
        string s;
        for(it i = ch1; i != ch2; ++i) { if(s.size())s += ','; s += chlist[*i-chbase].getname(); }
        if(s.size())s += ' ';
        
        list<Op> ops;
        for(it i = md1; i != md2; ++i)
            ops.push_back(mdlist[*i-mdbase]);
        
        s += strgroup(ops);
        
        vector<collitemtype> agents(ag1, ag2);
        for(unsigned a=0; a<agents.size(); ++a)agents[a] -= agbase;
        
        joblist.push_back(make_pair(s, agents));
    }
    
    void dump() const
    {
        typename joblisttype::const_iterator i;
        for(i=joblist.begin(); i!=joblist.end(); ++i)
        {
            printf("# Agents:");
            typename vector<collitemtype>::const_iterator j;
            for(j=i->second.begin(); j!=i->second.end(); ++j)
            {
                const Agent &a = aglist[*j];
                printf(" ");
                a.dump();
            }
            printf("\n# Job: \"%s\"\n# \n", i->first.c_str());
        }
    }
    
    void distribute()
    {
#ifdef ELEMLIST_DEBUG
        dump();
#endif
        typedef typename joblisttype::const_iterator tit;
        typedef unsigned combtype;
        
        combtype combcount=1;        
        for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            combcount *= i->second.size();
        
        vector<unsigned> loads(aglist.Size());
        for(unsigned b=0; b<loads.size(); ++b)
            loads[b] = aglist[b].getload();
        
        combtype bestcomb = 0;
        unsigned bestload = 0, leastload = 0, relying=0;
        for(combtype a=0; a<combcount; ++a)
        {
            vector<unsigned> tmploads = loads;
            combtype c=a;
            
            /* See how much load would this choice generate */
            /* Cumulate all agents' load */
            
            for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            {
                unsigned mod = i->second.size();
                unsigned agnum = c%mod; c/=mod;
                ++tmploads[i->second[agnum]];
            }
            
            /* Now see how much there's load totally */
            
            c=a;
            unsigned loadsum=0, relysum = 0;
            for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            {
                unsigned mod = i->second.size();
                unsigned agnum = c%mod; c/=mod;
                
                unsigned load = tmploads[i->second[agnum]];
                loadsum += load * load;
                relysum += aglist[i->second[agnum]].getrely();
            }

            unsigned maxload = 0;
            for(unsigned b=0; b<tmploads.size(); ++b)
                if(tmploads[b] > maxload)
                    maxload = tmploads[b];
            
#ifdef ELEMLIST_DEBUG
            printf("# Choice %u: Load would be %u (%u)", a, maxload, loadsum);
#endif
            if(maxload < bestload || !a)
            {
                bestload = maxload;
            Good:
                relying = relysum;
                leastload= loadsum;
                bestcomb = a; 
#ifdef ELEMLIST_DEBUG
                printf(" - using");
#endif
            }
            else if(maxload == bestload && (loadsum < leastload || relysum > relying))
                goto Good;
            
#ifdef ELEMLIST_DEBUG
            printf("\n");
            for(unsigned b=0; b<tmploads.size(); ++b)printf("%u ", tmploads[b]);
            printf("\n");
#endif
        }
        
        for(tit i=joblist.begin(); i!=joblist.end(); ++i)
        {
            unsigned mod = i->second.size();
            unsigned agnum = bestcomb%mod; bestcomb/=mod;
            
            printf("%c %s %s\n", mdlist[0].cmdchar(), aglist[i->second[agnum]].getname().c_str(), i->first.c_str());
        }
    }
};

template<class Op>
void ActMap<Op>::Distribute()
{
#ifdef ELEMLIST_DEBUG
    typedef char collitemtype;
    collitemtype aBase = 'A';
    collitemtype cBase = 'A' + As();
    collitemtype mBase = 'A' + As() + Cs();
    
    printf("# Agents: ");
    for(unsigned a=0; a<As(); ++a) { printf("%c(", (collitemtype)(a+aBase)); Agents[a].dump(); printf(") "); }
    printf("\n# Channels: ");
    for(unsigned c=0; c<Cs(); ++c) { printf("%c(", (collitemtype)(c+cBase)); Channels[c].dump(); printf(") "); }
    printf("\n# Ops: ");
    for(unsigned m=0; m<Ms(); ++m) { printf("%c(", (collitemtype)(m+mBase)); Ops[m].dump(); printf(") "); }
    printf("\n");
#else
    typedef unsigned collitemtype;
    collitemtype aBase=0, cBase=0, mBase=0;
#endif

    /* This task will be done completely differently. */
    if(Ops.Size() > 0 && Ops[0].cmdchar() == 'O')
    {
        typedef pair<unsigned, unsigned> operation;
        typedef std::map<operation, vector<unsigned> > operationmap;
        
        operationmap operations;
        
        for(unsigned m=0; m<Ms(); ++m)
            for(unsigned a=0; a<As(); ++a)
                for(unsigned c=0; c<Cs(); ++c)
                    if(map[MdInd(a,c,m)])
                        operations[make_pair(c,m)].push_back(a);
        
        vector<unsigned> loads(Agents.Size());
        for(unsigned b=0; b<loads.size(); ++b)
            loads[b] = Agents[b].gethostload();
        
        operationmap::const_iterator i;
        for(i=operations.begin(); i!=operations.end(); ++i)
        {
            /* For each operation, pick up the agent to do it. */
            /* Currently assume that the jobs are equal in complexity. */
            /* Jobs are not groupable, so this is quite easy. */
            
            unsigned leastload = 0x7FFFFFFF, maxrely = 0, good = 0;
            for(unsigned a=0; a<i->second.size(); ++a)
            {
                unsigned agnum = i->second[a];
                unsigned thisload = loads[agnum];
                unsigned thisrely = Agents[agnum].getrely();
                if(thisload < leastload)
                {
                    leastload = thisload;
                    maxrely = thisrely;
                    good = agnum;
                }
                else if(thisload == leastload && thisrely > maxrely)
                {
                    maxrely = thisrely;
                    good = agnum;
                }
            }
            
            //unsigned chnum = i->first.first;
            unsigned mdnum = i->first.second;
            ++loads[good];
            printf("%c %s %s\n",
                Ops[mdnum].cmdchar(),
                Agents[good].getname().c_str(),
                //Channels[chnum].getname().c_str(),
                Ops[mdnum].getparm().c_str()
                  );
        }
    }
    else
    {
        typedef elemlist<3, collitemtype> colltype;
        colltype coll;
        
        for(unsigned m=0; m<Ms(); ++m)
            for(unsigned a=0; a<As(); ++a)
                for(unsigned c=0; c<Cs(); ++c)
                    if(map[MdInd(a,c,m)])
                    {
                        colltype::Elem tmp;
                        tmp.insert(0, (collitemtype)(a+aBase));
                        tmp.insert(1, (collitemtype)(c+cBase));
                        tmp.insert(2, (collitemtype)(m+mBase));
                        coll.insert(tmp);
                    }

#ifdef ELEMLIST_DEBUG
        printf("# Unoptimized: ");
        coll.dump();
        printf("\n");
#endif

        coll.optimize();
#ifdef ELEMLIST_DEBUG
        printf("# Optimized: ");
        coll.dump();
        printf("\n");
#endif

        coll.expand3(Group3test<collitemtype,Op> (mBase, Ops));
#ifdef ELEMLIST_DEBUG
        printf("# Expanded: ");
        coll.dump();
        printf("\n");
#endif
        coll.crop();
#ifdef ELEMLIST_DEBUG
        printf("# Cropped: ");
        coll.dump();
        printf("\n");
#endif
        
        JobList<collitemtype, Op> jobs(
          aBase,cBase,mBase,
          Agents,Channels,Ops);
        coll.format(jobs);
        
#ifdef ELEMLIST_DEBUG
        printf("# ");
        coll.dump();
        printf("\n");
#endif

        jobs.distribute();
    }
}
template<class Op>
void ActMap<Op>::SetChanOp(const Channel &ch, const Op &md)
{
    unsigned mi = OpIndex(md);
    unsigned ci = ChannelIndex(ch);
    bool got = DoMarkOp(ci, mi);
    if(!got)
    {
        if(md.cmdchar() != 'G'  //msg
        && md.cmdchar() != 'O'  //functioncall
        && md.cmdchar() != 'N'  //notice
        && md.cmdchar() != 'Q'  //whois
                              ) // these functions do not need +o
        {
            printf("W ");
            ch.dump();
            printf(" %c ", md.cmdchar());
            md.dump(); 
            printf("\n");
        }
    }
}
