#include <stdio.h>
#include <string.h>
#include <string>
#include <vector>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sys/wait.h>
#include <fcntl.h>

#define DEBUG 0

/*
 This is a wrapper program for modecombiner, used by BisqBot.
 Copyright (C) 1992,2002 Bisqwit (http://iki.fi/bisqwit/)
 
 It is called like this:
 
 start runmodecombiner
 for each agent
   print "A %s", agent name
   for each agent's channel
     print "C %s %s", channel access code, channel name
   print "J %d", job count
   print "P %d", priority
   print "R 0 %s", hostname
 for each host
   print "H %s", hostname
   print "L %d", host load
 for each task
   print "D %c %s", task type, task
 print "Q"
 wait for results.
 
 The purpose of the program is to log all modecombiner-calls and to
 supress the useless modecombiner parameters (like useless agents etc).
*/

using namespace std;

class Channel
{
public:
    char State;  // 3=op, 2=voice, 1=nothing
    string Name;
    bool Needed;
    
    Channel() : Needed(false)
    {
    }
};
class Host
{
public:
    string Name;
    int BusyCount;
    
    Host()
    {
    }
    
    void Clear()
    {
        Name = "";
        BusyCount = 0;
    }
};
static bool StrEqual(const string &a, const string &b)
{
    if(a.size() != b.size())return false;
    for(unsigned c=0; c<a.size(); ++c)
        if(toupper(a[c]) != toupper(b[c]))return false;
    return true;
}
class Agent
{
public:
    string Name;
    int BusyCount;
    int Priority;
    vector<Channel> Channels;
    bool Needed;
    int fd;
    string Host;

    Agent()
    {
    }

    void Clear()
    {
        Channels.clear();
        Priority = BusyCount = 0;
        Name = "";
        Needed = false;
        fd = -1;
        Host = "";
    }
    bool CanHelp(char Operation, const string &ChName)
    {
        char RequiredState = '1';

        if(Operation == 'M' || Operation == 'K' || Operation == 'T')
            RequiredState = '3';
        else
            RequiredState = '1';
        
#if DEBUG
        fprintf(stderr, "I am '%s', can I help with %c on '%s', Required State = '%c'?\n",
            Name.c_str(), Operation, ChName.c_str(), RequiredState);
#endif
        
        for(unsigned a=0; a<Channels.size(); ++a)
        {
            Channel &c = Channels[a];
#if DEBUG
            fprintf(stderr, "- I have '%s' (state %c)", c.Name.c_str(), c.State);
#endif
            if(c.State >= RequiredState
            && StrEqual(c.Name, ChName))
            {
#if DEBUG
                fprintf(stderr, " - That will do.\n");
#endif
                c.Needed = true;
                return true;
            }
#if DEBUG
            fprintf(stderr, " - That does not help.\n");
#endif
        }
#if DEBUG
        fprintf(stderr, "No deal.\n");
#endif
        return false;
    }
};

int main(void)
{
    char Buf[2048];
    vector<string> Jobs;
    vector<Agent> Agents;
    vector<Host> Hosts;
    
    Agent ag;
    Host ho;
    char Operation = 0;
    
    bool AgentPending = false;
    bool HostPending = false;
    
    /* Get the command list from ircII */
    while((fgets(Buf, sizeof Buf, stdin)))
    {
#if DEBUG
        fprintf(stderr, "%s", Buf);
#endif
        
        if(Buf[0] == 'Q')break;
        strtok(Buf, "\n");
        switch(Buf[0])
        {
            case 'A': 
            {
                if(AgentPending) { Agents.push_back(ag); }
                ag.Clear();
                ag.Name = Buf+2;
                AgentPending = true;
                break;
            }
            case 'C':
            {
                Channel tmp;
                tmp.State = Buf[2];
                tmp.Name = Buf+4;
                tmp.Needed = false;
                ag.Channels.push_back(tmp);
                break;
            }
            case 'J':
            {
                ag.BusyCount = atoi(Buf+2);
                break;
            }
            case 'P':
            {
                ag.Priority = atoi(Buf+2);
                break;
            }
            case 'R':
            {
                ag.fd = atoi(Buf+2);
                const char *s = Buf+2;
                while(*s!=' ' && *s)++s;
                while(*s==' ')++s;
                ag.Host = s;
                break;
            }
            case 'H':
            {
                if(HostPending) { Hosts.push_back(ho); }
                ho.Clear();
                ho.Name = Buf+2;
                HostPending = true;
                break;
            }
            case 'L':
            {
                ho.BusyCount = atoi(Buf+2);
                break;
            }
            case 'D':
            {
                Operation = Buf[2];
                Jobs.push_back(Buf+4);
                break;
            }
            default:
                break;
        }
    }

    if(AgentPending) { Agents.push_back(ag); }
    if(HostPending) { Hosts.push_back(ho); }
    
    /* Sort out the channel list and see which agents prove useful */
    for(unsigned a=0; a<Jobs.size(); ++a)
    {
        string Job = Jobs[a];
        unsigned a = Job.find(' ');
        if(a == Job.npos)continue;
        Job = Job.substr(0, a);
        for(a=0; a<Job.size() ;)
        {
            unsigned b = Job.find(',', a);
            if(b == Job.npos)b = Job.size();
            
            string Channel = Job.substr(a, b-a);
            
            if(Channel[0] != '#'
            && Channel[0] != '&'
            && Channel[0] != '!'
            && Channel[0] != '&'
            && Channel[0] != '+')
            {
                /* Not a channel! Then every agent can be useful. */
                for(unsigned c=0; c<Agents.size(); ++c)
                    Agents[c].Needed = true;
            }
            else
            {
                for(unsigned c=0; c<Agents.size(); ++c)
                    if(Agents[c].CanHelp(Operation, Channel))
                    {
                        Agents[c].Needed = true;
                    }
            }
            a = b+1;
        }
    }
    
    /* Build the command line */
    vector<const char *> Argv;
    Argv.push_back("modecombiner");
    
    for(unsigned b=0; b<Agents.size(); ++b)
    {
        if(!Agents[b].Needed)continue;
        
        string AgentDef = Agents[b].Name;
        sprintf(Buf, ":%d:%d", Agents[b].BusyCount, Agents[b].Priority);
        AgentDef += Buf;
        
        for(unsigned a=0; a<Hosts.size(); ++a)
            if(Hosts[a].Name == Agents[b].Host)
            {
                sprintf(Buf, ":%d", Hosts[a].BusyCount);
                AgentDef += Buf;
                break;
            }

        {char *s = (char *)malloc(AgentDef.size() + 1);
        if(s)
        {
            strcpy(s, AgentDef.c_str());
            Argv.push_back(s);
        }}
        
        AgentDef = "";
        
        /* Find out if there's a nick specified:   *
         * every agent can be useful in that case. */
        for(unsigned a=0; a<Jobs.size(); ++a)
        {
            string Job = Jobs[a];
            unsigned a = Job.find(' ');
            if(a == Job.npos)continue;
            Job = Job.substr(0, a);
            for(a=0; a<Job.size() ;)
            {
                unsigned b = Job.find(',', a);
                if(b == Job.npos)b = Job.size();
                
                string Channel = Job.substr(a, b-a);
                
                if(Channel[0] != '#'
                && Channel[0] != '&'
                && Channel[0] != '!'
                && Channel[0] != '&'
                && Channel[0] != '+')
                {
                   if(AgentDef.size())AgentDef += ' ';
                   AgentDef += Jobs[a];
                }
                a = b+1;
            }
        }

        for(unsigned a=0; a<Agents[b].Channels.size(); ++a)
            if(Agents[b].Channels[a].Needed)
            {
                if(AgentDef.size())AgentDef += ' ';
                AgentDef += Agents[b].Channels[a].Name;
            }

        {char *s = (char *)malloc(AgentDef.size() + 1);
        if(s)
        {
            strcpy(s, AgentDef.c_str());
            Argv.push_back(s);
        }}
    }
    
    {char *s = (char *)malloc(3);
     if(s)
     {
         s[0] = '/';
         s[1] = Operation;
         s[2] = '\0';
         Argv.push_back(s);
     }
    }
    
    for(unsigned a=0; a<Jobs.size(); ++a)
    {
        char *s = (char *)malloc(Jobs[a].size() + 1);
        if(s)
        {
            strcpy(s, Jobs[a].c_str());
            Argv.push_back(s);
        }
    }
        
    Argv.push_back(NULL);
    
    string Command = "";
    for(unsigned a=0; a<Argv.size(); ++a)
    {
        if(!Argv[a])continue;
        if(a>0)Command += ' ';
        if(strchr(Argv[a], ' ') || Argv[a][0]=='#' || Argv[a][0]=='$')Command += '\'';
        Command += Argv[a];
        if(strchr(Argv[a], ' ') || Argv[a][0]=='#' || Argv[a][0]=='$')Command += '\'';
    }
    
    fflush(stdout);
    fflush(stderr);
    
    int nullfd = open("/dev/null", O_RDWR);
    /* Tehdn pipe, jotta parentin tappaminen tappaisi mys childin */
    int pip[2]; pipe(pip);
    int pid = vfork();
    
    if(!pid)
    {
    	dup2(pip[1], 1);
    	dup2(nullfd, 0);
    	close(pip[0]); close(pip[1]); close(nullfd);
    
        /* Switch to modecombiner */
        execvp(Argv[0], const_cast<char *const *> (&Argv[0]));
    }
    dup2(pip[0], 0);
    dup2(nullfd, 2);
    close(pip[0]); close(pip[1]); close(nullfd);
    
    string Output;
    for(;;)
    {
    	char Buf[4096];
    	
    	int c = read(0, Buf, sizeof Buf);
    	if(c <= 0)break;
    	
    	Output += string(Buf, c);
    }
    while(wait(NULL) > 0);
    
    fwrite(Output.data(), 1, Output.size(), stdout);
    fflush(stdout);
    
    /* Log the command */
    FILE *fp = fopen("volatile/execs.log", "a");
    if(!fp) fp = fopen("volatile/execs.log", "w");
    if(!fp) fp = fopen("execs.log", "a");
    if(!fp) fp = fopen("execs.log", "w");
    if(fp)
    {
        time_t t = time(NULL);
        struct tm *TM = localtime(&t);
        strftime(Buf, sizeof(Buf), "%Y-%m%d-%H%M.%S", TM);
        fprintf(fp, "%s %s\n", Buf, Command.c_str());
        fclose(fp);
    }
    
    return -1;
}

/*

EI NIN:
*** Sending signal SIGKILL (9) to process 0: runmodecombiner
2259AgentLog: For mode M: Killing process MODECOMM1018468775 and combining
2259AgentLog:   [#pkmn -v Unka]
2259AgentLog: And
2259AgentLog:   [#pkmn -v MafiaNine]
2259AgentLog: Resulting set:
2259AgentLog:   [#pkmn -v MafiaNine]
2259AgentLog:   [#pkmn -v Unka]
Received: MODECOMM1018468775 M Bisqwit #pkmn -v MafiaNine
Purged bs.opkeep[M] and bs.opkeep.MODECOMM1018468775
Bisqwit:dcc raw 6 stc .agentcmd //mode #pkmn -v MafiaNine
Received: MODECOMM1018468777 M Bisqwit #pkmn -vv MafiaNine Unka
Purged bs.opkeep[M] and bs.opkeep.MODECOMM1018468777
Bisqwit:dcc raw 6 stc .agentcmd //mode #pkmn -vv MafiaNine Unka




*/
